# For tighter subclass inheritance
from abc import ABCMeta, abstractmethod

# To allow for precise type hinting to the humans
from typing import ClassVar, Collection, Dict, Tuple, Union


class AbstractParameter(metaclass=ABCMeta):
    """
    An abstract representation of a cruise-control parameter.

    Parameters are the key=value mappings that follow the endpoint in a URL.
    """
    # A string, like 'brokerid', that cannot be supplied at instantiation.
    #
    # Concrete child classes of Parameter are expected to supply a class-level
    # value for this.
    #
    # This property is meant to correspond to the parameters listed at
    # https://github.com/linkedin/cruise-control/tree/master/cruise-control/src/main/java/com/linkedin/kafka/cruisecontrol/servlet/parameters
    name: ClassVar[str] = None
    description: ClassVar[str] = None
    # Define a convenience data structure to help in programmatically building CLIs
    argparse_properties: ClassVar[Dict[str, Union[Tuple[str], Dict[str, str]]]] = {
        'args': (),
        'kwargs': {}
    }

    def __init__(self, value: Union[bool, int, float, str] = None):
        # The value of this parameter, like '5', '123,234,345', or 'True'
        #
        # This value must be supplied at instantiation.
        #
        # Concrete child classes of Parameter can define an expected type
        # for the value property.
        self.value = value

        # Check that the supplied value is not incorrect.
        self.validate_value()

    def __hash__(self):
        # The (name, value) tuple should uniquely define a Parameter
        return hash((self.name, self.value))

    @abstractmethod
    def validate_value(self):
        """
        This method provides further validation on the supplied value.

        If the supplied value is invalid for this parameter, raises ValueError.
        """
        pass


class MinValidPartitionRatioParameter(AbstractParameter):
    """min_valid_partition_ratio=[min_valid_partition_ratio]"""
    name = 'min_valid_partition_ratio'
    description = 'The minimum required ratio of monitored topics [0-1]'
    argparse_properties = {
        'args': ('--min-valid-partition-ratio',),
        'kwargs': dict(help=description, metavar='R', type=float)
    }

    def __init__(self, value: float):
        AbstractParameter.__init__(self, value)

    def validate_value(self):
        if type(self.value) != float and type(self.value) != int:
            raise ValueError(f"{self.value} must be a float or an integer")
        if self.value < 0 or self.value > 1:
            raise ValueError(f"{self.value} must be between 0 and 1, inclusive")


class PartitionParameter(AbstractParameter):
    """partition=[partition/start_partition-end_partition]"""
    name = 'partition'
    description = "The partition or [start]-[end] partitions to return"
    argparse_properties = {
        'args': ('--partition', '--partitions'),
        'kwargs': dict(help=description, metavar='PARTITION_OR_RANGE')
    }

    def __init__(self, value: Union[int, Tuple[int, int], str]):
        if isinstance(value, Collection):
            value = '-'.join(value)
        AbstractParameter.__init__(self, value)

    def validate_value(self):
        if type(self.value) == int:
            pass
        elif type(self.value) == str:
            if len(self.value.split('-')) > 2:
                raise ValueError(f"{self.value} must contain either 1 integer, or 2 '-'-separated integers")
        else:
            raise ValueError(f"{self.value} must either be a string, an integer, or a Collection of two integers")


class ReasonParameter(AbstractParameter):
    """reason=[reason-for-pause]"""
    name = 'reason'
    description = 'The reason for pausing or resuming the sampling'
    argparse_properties = {
        'args': ('--reason',),
        'kwargs': dict(help=description, metavar='REASON', type=str)
    }

    def validate_value(self):
        """
        The validation strategy here is to avoid a Bobby Tables situation
        by ensuring that the provided reason does not contain any
        RFC-3986 "reserved" characters.

        See https://tools.ietf.org/html/rfc3986#section-2.2
        :return:
        """
        gen_delims = set(":/?#[]@")
        sub_delims = set("!$&'()*+,;=")
        reserved_html_characters = gen_delims | sub_delims
        for character in self.value:
            if character in reserved_html_characters:
                raise ValueError(f"\"{character}\" is a reserved HTML character"
                                 " and cannot be transmitted correctly to cruise-control"
                                 " as part of the reason= parameter")



from typing import Set, Union, Collection  # noqa

# To allow for warnings that don't break the program.
#
# Specifically, we don't want this client to break forward compatibility with
# new cruise-control features and parameter choices.
#
# Accordingly, for enums, we want to warn, but not raise an Exception, when
# a provided value is outside of that enum
from warnings import warn

from cruisecontrolclient.client.CCParameter.Parameter import AbstractParameter


class AbstractSetOfChoicesParameter(AbstractParameter):
    lowercase_set_of_choices: Set[str] = set()

    def __init__(self, value: Union[str, Collection[str]]):
        if isinstance(value, Collection) and type(value) != str:
            value = ','.join(value)
        AbstractParameter.__init__(self, value)

    def validate_value(self):
        if type(self.value) == str:
            for i in self.value.split(','):
                # Assume that cruise-control uses case-insensitive enums
                if i.lower() not in self.lowercase_set_of_choices:
                    warn(f"{i} is not in the valid set of choices")
        else:
            raise ValueError(f"{self.value} must be either a string or a Collection")


class DisableSelfHealingForParameter(AbstractSetOfChoicesParameter):
    """disable_self_healing_for=[Set-of-{@link AnomalyType}]"""
    lowercase_set_of_choices = {'goal_violation', 'metric_anomaly', 'broker_failure'}
    name = 'disable_self_healing_for'
    description = "The anomaly detectors for which to disable self-healing"
    argparse_properties = {
        'args': ('--disable-self-healing-for', '--disable-self-healing'),
        # nargs='+' allows for multiple of the set to be specified
        'kwargs': dict(help=description, metavar='ANOMALY_DETECTOR', choices=lowercase_set_of_choices, nargs='+')

    }


class EnableSelfHealingForParameter(AbstractSetOfChoicesParameter):
    """enable_self_healing_for=[Set-of-{@link AnomalyType}]"""
    lowercase_set_of_choices = {'goal_violation', 'metric_anomaly', 'broker_failure'}
    name = 'enable_self_healing_for'
    description = "The anomaly detectors for which to enable self-healing"
    argparse_properties = {
        'args': ('--enable-self-healing-for', '--enable-self-healing'),
        # nargs='+' allows for multiple of the set to be specified
        'kwargs': dict(help=description, metavar='ANOMALY_DETECTOR', choices=lowercase_set_of_choices, nargs='+')
    }


class ResourceParameter(AbstractSetOfChoicesParameter):
    """resource=[RESOURCE]"""
    lowercase_set_of_choices = {"cpu", "networkinbound", "networkoutbound", "disk"}
    name = 'resource'
    description = "The host and broker-level resource by which to sort the cruise-control response"
    argparse_properties = {
        'args': ('--resource',),
        'kwargs': dict(help=description, metavar='RESOURCE', choices=lowercase_set_of_choices)
    }


class SubstatesParameter(AbstractSetOfChoicesParameter):
    """substates=[SUBSTATES]"""
    lowercase_set_of_choices = {'executor', 'analyzer', 'monitor', 'anomaly_detector'}
    name = 'substates'
    description = "The substates for which to retrieve state from cruise-control"
    argparse_properties = {
        'args': ('--substate', '--substates'),
        # nargs='+' allows for multiple of the set to be specified
        'kwargs': dict(help=description, metavar='SUBSTATE', choices=lowercase_set_of_choices, nargs='+')
    }



# Flatten the imports so that packages above just need:
# from CCParameter import {SomeConcreteParameterClass}
from cruisecontrolclient.client.CCParameter.BooleanParameter import *  # noqa
from cruisecontrolclient.client.CCParameter.CommaSeparatedParameter import *  # noqa
from cruisecontrolclient.client.CCParameter.NonNegativeIntegerParameter import *  # noqa
from cruisecontrolclient.client.CCParameter.Parameter import *  # noqa
from cruisecontrolclient.client.CCParameter.PositiveIntegerParameter import *  # noqa
from cruisecontrolclient.client.CCParameter.RegularExpressionParameter import *  # noqa
from cruisecontrolclient.client.CCParameter.SetOfChoicesParameter import *  # noqa
from cruisecontrolclient.client.CCParameter.TimeStampParameter import *  # noqa



# Copyright 2019 LinkedIn Corp. Licensed under the BSD 2-Clause License (the "License").
# See License in the project root for license information.
import sys


def print_error(*args, sep=' ', end='\n'):
    """
    Like the print() builtin, but presumes printing to stderr, rather than stdout
    :param sep:   string inserted between values, default a space.
    :param end:   string appended after the last value, default a newline.
    :return:
    """
    print(*args, sep=sep, end=end, file=sys.stderr)



# Copyright 2019 LinkedIn Corp. Licensed under the BSD 2-Clause License (the "License").
# See License in the project root for license information.

# To be able to instantiate Endpoint objects
import cruisecontrolclient.client.Endpoint as Endpoint

# To be able to make more precise type hints
from typing import Dict, Tuple, Type  # noqa


class ExecutionContext:
    # Define the in-order available endpoints for programmatically building the argparse CLI
    available_endpoints: Tuple[Type[Endpoint.AbstractEndpoint]] = (
        Endpoint.AddBrokerEndpoint,
        Endpoint.AdminEndpoint,
        Endpoint.BootstrapEndpoint,
        Endpoint.DemoteBrokerEndpoint,
        Endpoint.FixOfflineReplicasEndpoint,
        Endpoint.KafkaClusterStateEndpoint,
        Endpoint.LoadEndpoint,
        Endpoint.PartitionLoadEndpoint,
        Endpoint.PauseSamplingEndpoint,
        Endpoint.ProposalsEndpoint,
        Endpoint.RebalanceEndpoint,
        Endpoint.RemoveBrokerEndpoint,
        Endpoint.ResumeSamplingEndpoint,
        Endpoint.ReviewBoardEndpoint,
        Endpoint.ReviewEndpoint,
        Endpoint.StateEndpoint,
        Endpoint.StopProposalExecutionEndpoint,
        Endpoint.TrainEndpoint,
        Endpoint.TopicConfigurationEndpoint,
        Endpoint.UserTasksEndpoint
    )

    # A mapping from the names of the subparsers to the Endpoint object they should instantiate.
    #
    # Like:
    # {"add_broker": AddBrokerEndpoint,
    #  "add_brokers": AddBrokerEndpoint,
    #  "admin": AdminEndpoint,
    #  "bootstrap": BootstrapEndpoint,
    #  "demote_broker": DemoteBrokerEndpoint,
    #  "demote_brokers": DemoteBrokerEndpoint,
    #  "fix_offline_replicas": FixOfflineReplicasEndpoint,
    #  "kafka_cluster_state": KafkaClusterStateEndpoint,
    #  "load": LoadEndpoint,
    #  "partition_load": PartitionLoadEndpoint,
    #  "pause_sampling": PauseSamplingEndpoint,
    #  "proposals": ProposalsEndpoint,
    #  "rebalance": RebalanceEndpoint,
    #  "remove_broker": RemoveBrokerEndpoint,
    #  "remove_brokers": RemoveBrokerEndpoint,
    #  "resume_sampling": ResumeSamplingEndpoint,
    #  "state": StateEndpoint,
    #  "stop_proposal_execution": StopProposalExecutionEndpoint,
    #  "stop": StopProposalExecutionEndpoint,
    #  "train": TrainEndpoint,
    #  "user_task": UserTasksEndpoint,
    #  "user_tasks": UserTasksEndpoint}
    dest_to_Endpoint: Dict[str, Type[Endpoint.AbstractEndpoint]] = {}
    available_parameter_set = set()
    for endpoint in available_endpoints:
        dest_to_Endpoint[endpoint.name] = endpoint
        if 'aliases' in endpoint.argparse_properties['kwargs']:
            endpoint_aliases = endpoint.argparse_properties['kwargs']['aliases']
            for alias in endpoint_aliases:
                dest_to_Endpoint[alias] = endpoint
        for parameter in endpoint.available_Parameters:
            available_parameter_set.add(parameter)

    # A mapping from the names of the argparse parameters to their cruise-control parameter name
    #
    # Like:
    # {'allow_capacity_estimation': 'allow_capacity_estimation',
    #  'brokers': 'brokerid',
    #  'clearmetrics': 'clearmetrics',
    #  'client_id': 'client_ids',
    #  'concurrency': 'concurrent_partition_movements_per_broker',
    #  'data_from': 'data_from',
    #  'disable_self_healing_for': 'disable_self_healing_for',
    #  'dry_run': 'dryRun',
    #  'enable_self_healing_for': 'enable_self_healing_for',
    #  'end_timestamp': 'end',
    #  'endpoint': 'endpoints',
    #  'exclude_follower_demotion': 'exclude_follower_demotion',
    #  'exclude_recently_demoted_brokers': 'exclude_recently_demoted_brokers',
    #  'exclude_recently_removed_brokers': 'exclude_recently_removed_brokers',
    #  'excluded_topics': 'excluded_topics',
    #  'goals': 'goals',
    #  'ignore_proposal_cache': 'ignore_proposal_cache',
    #  'leader_concurrency': 'concurrent_leader_movements',
    #  'max_load': 'max_load',
    #  'min_valid_partition_ratio': 'min_valid_partition_ratio',
    #  'number_of_entries_to_show': 'entries',
    #  'partition': 'partition',
    #  'reason': 'reason',
    #  'resource': 'resource',
    #  'skip_hard_goal_check': 'skip_hard_goal_check',
    #  'skip_urp_demotion': 'skip_urp_demotion',
    #  'start_timestamp': 'start',
    #  'strategies': 'replica_movement_strategies',
    #  'substate': 'substates',
    #  'super_verbose': 'super_verbose',
    #  'text': 'json',
    #  'throttle_removed_broker': 'throttle_removed_broker',
    #  'timestamp': 'time',
    #  'topic': 'topic',
    #  'types': 'types',
    #  'use_ready_default_goals': 'use_ready_default_goals',
    #  'user_task_ids': 'user_task_ids',
    #  'verbose': 'verbose'}
    flag_to_parameter_name: Dict[str, str] = {}
    for parameter in available_parameter_set:
        argparse_parameter_name = None
        # argparse names this parameter's flag after the first string in 'args'
        for possible_argparse_name in parameter.argparse_properties['args']:
            # argparse chooses flag names only from the --flags
            if not possible_argparse_name.startswith('-'):
                argparse_parameter_name = possible_argparse_name.replace('-', '_')
            elif not possible_argparse_name.startswith('--'):
                continue
            else:
                argparse_parameter_name = possible_argparse_name.lstrip('-').replace('-', '_')
                break
        if argparse_parameter_name and argparse_parameter_name in flag_to_parameter_name:
            raise ValueError(f"Colliding parameter flags: {argparse_parameter_name}")
        else:
            flag_to_parameter_name[argparse_parameter_name] = parameter.name

    def __init__(self):
        # Define the argparse flags that cannot be used for endpoint parameters
        #
        # This helps the multi-stage argument parsing not to conflict with itself
        # during the different stages.
        #
        # To this will probably be added things like:
        #   'add_parameter'
        #   'endpoint_subparser'
        #   'socket_address'
        #   'remove_parameter'
        #
        # If you've added flags to the argument_parser, they should be added to
        # this set.
        self.non_parameter_flags = set()



# Copyright 2019 LinkedIn Corp. Licensed under the BSD 2-Clause License (the "License").
# See License in the project root for license information.

# For tighter subclass inheritance
from abc import ABCMeta

# To allow us to make use of the Parameter convenience class
import cruisecontrolclient.client.CCParameter as CCParameter

# To allow us to make more-precise type hints
from typing import Callable, ClassVar, Dict, List, Tuple, Union  # noqa

# To help us generate the right parameter string from a dict of parameters
from urllib.parse import urlencode


class AbstractEndpoint(metaclass=ABCMeta):
    """
    An abstract representation of a cruise-control endpoint.

    Note that this class also provides methods for returning a correctly-
    concatenated endpoint with parameters.
    """

    # A string, like 'load' or 'state'
    # Meant to match those enumerated Strings at:
    # https://github.com/linkedin/cruise-control/blob/master/cruise-control/src/main/java/com/linkedin/kafka/cruisecontrol/servlet/EndPoint.java#L15
    name: ClassVar[str]

    # A human-readable string that describes this endpoint
    description: ClassVar[str]

    # A string, like 'GET' or 'POST'
    http_method: ClassVar[str]

    # Whether this endpoint can mutate the kafka cluster via a proposal execution.
    #
    # Primarily, this flag is intended to signal to users of this client
    # when they may need to perform de-conflict checks on the targeted kafka cluster.
    can_execute_proposal: ClassVar[bool]

    # An ordered collection of the known Parameter classes that can be instantiated for this Endpoint.
    available_Parameters: ClassVar[Tuple[CCParameter.AbstractParameter]]

    # Define a convenience data structure to help in programmatically building CLIs
    argparse_properties: ClassVar[Dict[str, Union[Tuple[str], Dict[str, str]]]] = \
        {
            'args': (),
            'kwargs': {}
        }

    def __init__(self):
        # A mapping of 'parameter' strings
        self.parameter_name_to_available_Parameters: Dict[
            str, Callable[[Union[str, int, bool]], CCParameter.AbstractParameter]] = \
            {ap.name: ap for ap in self.available_Parameters}

        # Stores the instantiated Parameters for this Endpoint.
        #
        # As parameters are added via add_param, if their 'parameter'
        self.parameter_name_to_instantiated_Parameters: Dict[str, CCParameter.AbstractParameter] = {}

        # Stores the URL parameters for which there is no Parameter class defined.
        #
        # This is intended to future-proof against cruise-control adding new
        # parameters before this client has a chance to implement them.
        self.parameter_name_to_value: Dict[str, str] = {}

    def add_param(self, parameter_name: str, value: Union[str, int, bool]) -> None:
        """
        Adds this parameter to this Endpoint, overriding any previous value-definition
        for this parameter.

        If the supplied 'parameter' matches one of the known Parameters for this
        Endpoint, that Parameter will be instantiated and its value validated.
        If the supplied value is not valid, Parameter will raise a ValueError.

        If the supplied 'parameter' does not match one of the known Parameters
        for this Endpoint, that 'parameter': 'value' mapping will be added to
        self.parameter_name_to_value without validation.

        :param parameter_name:
        :param value:
        :return:
        """
        if parameter_name in self.parameter_name_to_available_Parameters:
            Parameter_to_instantiate = self.parameter_name_to_available_Parameters[parameter_name]
            self.parameter_name_to_instantiated_Parameters[parameter_name] = Parameter_to_instantiate(value)
        else:
            self.parameter_name_to_value[parameter_name] = value

    def get_value(self, parameter_name: str) -> Union[str, None]:
        """
        Returns value if this parameter exists in this endpoint.

        Returns None otherwise

        :param parameter_name:
        :return:
        """
        if parameter_name in self.parameter_name_to_instantiated_Parameters:
            return self.parameter_name_to_instantiated_Parameters[parameter_name].value
        elif parameter_name in self.parameter_name_to_value:
            return self.parameter_name_to_value[parameter_name]
        else:
            return None

    def has_param(self, parameter_name: str) -> bool:
        """
        Returns True if this endpoint already has this parameter, False otherwise.

        :param parameter_name:
        :return:
        """
        return parameter_name in self.parameter_name_to_instantiated_Parameters or parameter_name in self.parameter_name_to_value

    def remove_param(self, parameter_name: str) -> None:
        """
        Remove this parameter from this Endpoint.

        If the supplied 'parameter' matches one of the known Parameters for this
        Endpoint, that Parameter will be removed from self.parameter_name_to_instantiated_Parameters.

        If the supplied 'parameter' does not match one of the known Parameters
        for this Endpoint, that 'parameter': 'value' mapping will be removed from
        self.parameter_name_to_value, if present.

        :param parameter_name:
        :return:
        """
        if parameter_name in self.parameter_name_to_instantiated_Parameters:
            self.parameter_name_to_instantiated_Parameters.pop(parameter_name)
        elif parameter_name in self.parameter_name_to_value:
            self.parameter_name_to_value.pop(parameter_name)

    def compose_endpoint(self) -> str:
        """
        Returns a valid URL suffix of this endpoint and any parameters
        that have been defined for it.

        Note that the ordering of the parameters is not guaranteed.

        :return: A string like:
            'rebalance?dryrun=false&allow_capacity_estimation=false&concurrent_partition_movements_per_broker=5'
            'state'
            'stop_proposal_execution'
        """
        # All parameter:value mappings that have been specified for this endpoint.
        #
        # First, add the parameter=value pairs from the Parameters we could instantiate.
        # Next, add (or override) with the parameter=value pairs that were explicitly defined.
        combined_parameter_to_value = {}

        # Update our mapping with parameter: value pairs from Parameter objects
        if self.parameter_name_to_instantiated_Parameters:
            combined_parameter_to_value.update(
                {name: ip.value for name, ip in self.parameter_name_to_instantiated_Parameters.items()})
        # Update our mapping with parameter: value pairs that lack Parameter objects
        if self.parameter_name_to_value:
            combined_parameter_to_value.update(self.parameter_name_to_value)

        # If we have any mappings, urlencode them and return the full string
        if combined_parameter_to_value:
            return f"{self.name}?{urlencode(combined_parameter_to_value)}"
        # Otherwise, just return this Endpoint's name
        else:
            return self.name


class AddBrokerEndpoint(AbstractEndpoint):
    name = "add_broker"
    description = "Move partitions to the specified brokers, according to the specified goals"
    http_method = "POST"
    can_execute_proposal = True
    available_Parameters = (
        CCParameter.AllowCapacityEstimationParameter,
        CCParameter.BrokerIdParameter,
        CCParameter.ConcurrentLeaderMovementsParameter,
        CCParameter.ConcurrentPartitionMovementsPerBrokerParameter,
        CCParameter.DryRunParameter,
        CCParameter.ExcludeRecentlyDemotedBrokersParameter,
        CCParameter.ExcludeRecentlyRemovedBrokersParameter,
        CCParameter.ExcludedTopicsParameter,
        CCParameter.GoalsParameter,
        CCParameter.JSONParameter,
        CCParameter.ReviewIDParameter,
        CCParameter.ReplicaMovementStrategiesParameter,
        CCParameter.SkipHardGoalCheckParameter,
        CCParameter.ThrottleRemovedBrokerParameter,
        CCParameter.UseReadyDefaultGoalsParameter,
        CCParameter.VerboseParameter
    )
    argparse_properties = {
        'args': (name,),
        'kwargs': dict(aliases=['add_brokers', 'add-broker', 'add-brokers'], help=description)
    }

    def __init__(self, broker_ids: Union[str, List[str]]):
        AbstractEndpoint.__init__(self)
        self.add_param("brokerid", broker_ids)


class AdminEndpoint(AbstractEndpoint):
    name = "admin"
    description = "Used to change runtime configurations on the cruise-control server itself"
    http_method = "POST"
    can_execute_proposal = False
    available_Parameters = (
        CCParameter.ConcurrentLeaderMovementsParameter,
        CCParameter.ConcurrentPartitionMovementsPerBrokerParameter,
        CCParameter.DisableSelfHealingForParameter,
        CCParameter.EnableSelfHealingForParameter,
        CCParameter.JSONParameter,
        CCParameter.ReviewIDParameter,
    )
    argparse_properties = {
        'args': (name,),
        'kwargs': dict(help=description)
    }


class BootstrapEndpoint(AbstractEndpoint):
    name = "bootstrap"
    description = "Bootstrap the load monitor"
    http_method = "GET"
    can_execute_proposal = False
    available_Parameters = (
        CCParameter.ClearMetricsParameter,
        CCParameter.EndParameter,
        CCParameter.JSONParameter,
        CCParameter.StartParameter
    )
    argparse_properties = {
        'args': (name,),
        'kwargs': dict(help=description)
    }


class DemoteBrokerEndpoint(AbstractEndpoint):
    name = "demote_broker"
    description = "Remove leadership and preferred leadership from the specified brokers"
    http_method = "POST"
    can_execute_proposal = True
    available_Parameters = (
        CCParameter.AllowCapacityEstimationParameter,
        CCParameter.BrokerIdParameter,
        CCParameter.ConcurrentLeaderMovementsParameter,
        CCParameter.DryRunParameter,
        CCParameter.ExcludeFollowerDemotionParameter,
        CCParameter.ExcludeRecentlyDemotedBrokersParameter,
        CCParameter.JSONParameter,
        CCParameter.ReplicaMovementStrategiesParameter,
        CCParameter.ReviewIDParameter,
        CCParameter.SkipURPDemotionParameter,
        CCParameter.VerboseParameter
    )
    argparse_properties = {
        'args': (name,),
        'kwargs': dict(aliases=['demote_brokers', 'demote-broker', 'demote-brokers'], help=description)
    }

    def __init__(self, broker_ids: Union[str, List[str]]):
        AbstractEndpoint.__init__(self)
        self.add_param("brokerid", broker_ids)


class FixOfflineReplicasEndpoint(AbstractEndpoint):
    # Warning, this Endpoint is only supported in kafka 1.1 and above
    name = "fix_offline_replicas"
    description = "Fixes the offline replicas in the cluster (kafka 1.1+ only)"
    http_method = "POST"
    can_execute_proposal = True
    available_Parameters = (
        CCParameter.ReviewIDParameter,
    )
    argparse_properties = {
        'args': (name,),
        'kwargs': dict(aliases=[name.replace('_', '-')], help=description)
    }


class KafkaClusterStateEndpoint(AbstractEndpoint):
    name = "kafka_cluster_state"
    description = "Get under-replicated and offline partitions (and under MinISR partitions in kafka 2.0+)"
    http_method = "GET"
    can_execute_proposal = False
    available_Parameters = (
        CCParameter.JSONParameter,
        CCParameter.VerboseParameter
    )
    argparse_properties = {
        'args': (name,),
        'kwargs': dict(aliases=[name.replace('_', '-')], help=description)
    }


class LoadEndpoint(AbstractEndpoint):
    name = "load"
    description = "Get the load on each kafka broker"
    http_method = "GET"
    can_execute_proposal = False
    available_Parameters = (
        CCParameter.AllowCapacityEstimationParameter,
        CCParameter.JSONParameter,
        CCParameter.TimeParameter
    )
    argparse_properties = {
        'args': (name,),
        'kwargs': dict(help=description)
    }

    def __init__(self):
        AbstractEndpoint.__init__(self)


class PartitionLoadEndpoint(AbstractEndpoint):
    name = "partition_load"
    description = "Get the resource load for each partition"
    http_method = "GET"
    can_execute_proposal = False
    available_Parameters = (
        CCParameter.AllowCapacityEstimationParameter,
        CCParameter.EndParameter,
        CCParameter.EntriesParameter,
        CCParameter.JSONParameter,
        CCParameter.MaxLoadParameter,
        CCParameter.MinValidPartitionRatioParameter,
        CCParameter.PartitionParameter,
        CCParameter.ResourceParameter,
        CCParameter.StartParameter,
        CCParameter.TopicParameter
    )
    argparse_properties = {
        'args': (name,),
        'kwargs': dict(aliases=[name.replace('_', '-')], help=description)
    }


class PauseSamplingEndpoint(AbstractEndpoint):
    name = "pause_sampling"
    description = "Pause metrics load sampling"
    http_method = "POST"
    can_execute_proposal = False
    available_Parameters = (
        CCParameter.JSONParameter,
        CCParameter.ReasonParameter,
        CCParameter.ReviewIDParameter,
    )
    argparse_properties = {
        'args': (name,),
        'kwargs': dict(aliases=[name.replace('_', '-')], help=description)
    }


class ProposalsEndpoint(AbstractEndpoint):
    name = "proposals"
    description = "Get current proposals"
    http_method = "GET"
    can_execute_proposal = False
    available_Parameters = (
        CCParameter.AllowCapacityEstimationParameter,
        CCParameter.DataFromParameter,
        CCParameter.ExcludeRecentlyDemotedBrokersParameter,
        CCParameter.ExcludeRecentlyRemovedBrokersParameter,
        CCParameter.ExcludedTopicsParameter,
        CCParameter.GoalsParameter,
        CCParameter.IgnoreProposalCacheParameter,
        CCParameter.JSONParameter,
        CCParameter.UseReadyDefaultGoalsParameter,
        CCParameter.VerboseParameter
    )
    argparse_properties = {
        'args': (name,),
        'kwargs': dict(help=description)
    }


class RebalanceEndpoint(AbstractEndpoint):
    name = "rebalance"
    description = "Rebalance the partition distribution in the kafka cluster, according to the specified goals"
    http_method = "POST"
    can_execute_proposal = True
    available_Parameters = (
        CCParameter.AllowCapacityEstimationParameter,
        CCParameter.ConcurrentLeaderMovementsParameter,
        CCParameter.ConcurrentPartitionMovementsPerBrokerParameter,
        CCParameter.DestinationBrokerIdsParameter,
        CCParameter.DryRunParameter,
        CCParameter.ExcludeRecentlyDemotedBrokersParameter,
        CCParameter.ExcludeRecentlyRemovedBrokersParameter,
        CCParameter.ExcludedTopicsParameter,
        CCParameter.GoalsParameter,
        CCParameter.IgnoreProposalCacheParameter,
        CCParameter.JSONParameter,
        CCParameter.ReplicaMovementStrategiesParameter,
        CCParameter.ReviewIDParameter,
        CCParameter.SkipHardGoalCheckParameter,
        CCParameter.UseReadyDefaultGoalsParameter,
        CCParameter.VerboseParameter
    )
    argparse_properties = {
        'args': (name,),
        'kwargs': dict(help=description)
    }

    def __init__(self):
        AbstractEndpoint.__init__(self)


class RemoveBrokerEndpoint(AbstractEndpoint):
    name = "remove_broker"
    description = "Remove all partitions from the specified brokers, according to the specified goals"
    http_method = "POST"
    can_execute_proposal = True
    available_Parameters = (
        CCParameter.AllowCapacityEstimationParameter,
        CCParameter.BrokerIdParameter,
        CCParameter.ConcurrentLeaderMovementsParameter,
        CCParameter.ConcurrentPartitionMovementsPerBrokerParameter,
        CCParameter.DestinationBrokerIdsParameter,
        CCParameter.DryRunParameter,
        CCParameter.ExcludeRecentlyDemotedBrokersParameter,
        CCParameter.ExcludeRecentlyRemovedBrokersParameter,
        CCParameter.ExcludedTopicsParameter,
        CCParameter.GoalsParameter,
        CCParameter.JSONParameter,
        CCParameter.ReplicaMovementStrategiesParameter,
        CCParameter.ReviewIDParameter,
        CCParameter.SkipHardGoalCheckParameter,
        CCParameter.ThrottleRemovedBrokerParameter,
        CCParameter.UseReadyDefaultGoalsParameter,
        CCParameter.VerboseParameter
    )
    argparse_properties = {
        'args': (name,),
        'kwargs': dict(aliases=['remove_brokers', 'remove-broker', 'remove-brokers'], help=description)
    }

    def __init__(self, broker_ids: Union[str, List[str]]):
        AbstractEndpoint.__init__(self)
        self.add_param("brokerid", broker_ids)


class ResumeSamplingEndpoint(AbstractEndpoint):
    name = "resume_sampling"
    description = "Resume metrics load sampling"
    http_method = "POST"
    can_execute_proposal = False
    available_Parameters = {
        CCParameter.JSONParameter,
        CCParameter.ReasonParameter,
        CCParameter.ReviewIDParameter,
    }
    argparse_properties = {
        'args': (name,),
        'kwargs': dict(aliases=[name.replace('_', '-')], help=description)
    }


class ReviewEndpoint(AbstractEndpoint):
    name = "review"
    description = "Create, approve, or discard reviews"
    http_method = "POST"
    can_execute_proposal = False
    available_Parameters = (
        CCParameter.ApproveParameter,
        CCParameter.DiscardParameter,
        CCParameter.JSONParameter,
        CCParameter.ReasonParameter
    )
    argparse_properties = {
        'args': (name,),
        'kwargs': dict(help=description)
    }


class ReviewBoardEndpoint(AbstractEndpoint):
    name = "review_board"
    description = "View already-created reviews"
    http_method = "GET"
    can_execute_proposal = False
    available_Parameters = (
        CCParameter.JSONParameter,
        CCParameter.ReviewIDsParameter
    )
    argparse_properties = {
        'args': (name,),
        'kwargs': dict(aliases=[name.replace('_', '-')], help=description)
    }


class StateEndpoint(AbstractEndpoint):
    name = "state"
    description = "Get the state of cruise control"
    http_method = "GET"
    can_execute_proposal = False
    available_Parameters = (
        CCParameter.JSONParameter,
        CCParameter.SubstatesParameter,
        CCParameter.SuperVerboseParameter,
        CCParameter.VerboseParameter
    )
    argparse_properties = {
        'args': (name,),
        'kwargs': dict(help=description)
    }

    def __init__(self):
        AbstractEndpoint.__init__(self)
        # Note that we most often want the executor substate, so set that
        # as a default parameter
        self.add_param("substates", "executor")


class StopProposalExecutionEndpoint(AbstractEndpoint):
    name = "stop_proposal_execution"
    description = "Stop the currently-executing proposal"
    http_method = "POST"
    can_execute_proposal = False
    available_Parameters = (
        CCParameter.JSONParameter,
        CCParameter.ReviewIDParameter,
    )
    argparse_properties = {
        'args': (name,),
        'kwargs': dict(aliases=[name.replace('_', '-'), 'stop'], help=description)
    }


class TopicConfigurationEndpoint(AbstractEndpoint):
    name = "topic_configuration"
    description = "Update the configuration of the specified topics"
    http_method = "POST"
    can_execute_proposal = True
    available_Parameters = (
        CCParameter.JSONParameter,
        CCParameter.SkipRackAwarenessCheckParameter,
        CCParameter.ReplicationFactorParameter,
        CCParameter.ReviewIDParameter,
        CCParameter.TopicParameter
    )
    argparse_properties = {
        'args': (name,),
        'kwargs': dict(aliases=[name.replace('_', '-'), 'stop'], help=description)
    }


class TrainEndpoint(AbstractEndpoint):
    name = "train"
    description = "Train the linear regression model"
    http_method = "GET"
    can_execute_proposal = False
    available_Parameters = (
        CCParameter.EndParameter,
        CCParameter.JSONParameter,
        CCParameter.StartParameter
    )
    argparse_properties = {
        'args': (name,),
        'kwargs': dict(help=description)
    }


class UserTasksEndpoint(AbstractEndpoint):
    name = "user_tasks"
    description = "Get the recent user tasks from cruise control"
    http_method = "GET"
    can_execute_proposal = False
    available_Parameters = (
        CCParameter.ClientIdsParameter,
        CCParameter.EndpointsParameter,
        CCParameter.EntriesParameter,
        CCParameter.JSONParameter,
        CCParameter.TypesParameter,
        CCParameter.UserTaskIdsParameter
    )
    argparse_properties = {
        'args': (name,),
        'kwargs': dict(aliases=['user_task', 'user-tasks', 'user-task'], help=description)
    }



from cruisecontrolclient.client.CCParameter.Parameter import AbstractParameter


class AbstractPositiveIntegerParameter(AbstractParameter):
    def __init__(self, value: int):
        AbstractParameter.__init__(self, value)

    def validate_value(self):
        if type(self.value) != int:
            raise ValueError(f"{self.value} is not an integer value")
        elif self.value < 1:
            raise ValueError(f"{self.value} must be a positive integer")


class ConcurrentLeaderMovementsParameter(AbstractPositiveIntegerParameter):
    """concurrent_leader_movements=[POSITIVE-INTEGER]"""
    name = 'concurrent_leader_movements'
    description = 'The maximum number of concurrent leadership movements across the entire cluster'
    argparse_properties = {
        'args': ('--leader-concurrency', '--leadership-concurrency', '--concurrent-leader-movements'),
        'kwargs': dict(metavar='K', help=description, type=int)
    }


class ConcurrentPartitionMovementsPerBrokerParameter(AbstractPositiveIntegerParameter):
    """concurrent_partition_movements_per_broker=[POSITIVE-INTEGER]"""
    name = 'concurrent_partition_movements_per_broker'
    description = 'The maximum number of concurrent partition movements per broker'
    argparse_properties = {
        'args': ('--concurrency', '--concurrent-partition-movements-per-broker'),
        'kwargs': dict(metavar='K', help=description, type=int)
    }


class DataFromParameter(AbstractPositiveIntegerParameter):
    """data_from=[valid_windows/valid_partitions]"""
    name = 'data_from'
    description = "The number of valid [windows, partitions] from which to use data"
    argparse_properties = {
        'args': ('--data-from',),
        'kwargs': dict(metavar='K', help=description, type=int)
    }


class EntriesParameter(AbstractPositiveIntegerParameter):
    """entries=[number-of-entries-to-show]"""
    name = 'entries'
    description = 'The number of entries to show in the response'
    argparse_properties = {
        'args': ('--number-of-entries-to-show', '--num-entries'),
        'kwargs': dict(metavar='K', help=description, type=int)
    }


class ReplicationFactorParameter(AbstractPositiveIntegerParameter):
    """replication_factor=[target_replication_factor]"""
    name = 'replication_factor'
    description = 'The target replication factor to which the specified topics should be set'
    argparse_properties = {
        'args': ('--replication-factor',),
        'kwargs': dict(metavar='K', help=description, type=int)
    }



from typing import Union, List

from cruisecontrolclient.client.CCParameter.Parameter import AbstractParameter


class AbstractCommaSeparatedParameter(AbstractParameter):
    def __init__(self, value: Union[str, List[str]]):
        # Transform the list into a comma-separated string to store self.value
        # in the format that is closest to what cruise-control expects
        if type(value) == list:
            value = ','.join(value)
        AbstractParameter.__init__(self, value)

    def validate_value(self):
        if type(self.value) != str:
            raise ValueError(f"{self.value} is not a string value")


class ApproveParameter(AbstractCommaSeparatedParameter):
    """approve=[id1,id2,...]"""
    name = 'approve'
    description = "The review IDs to approve"
    argparse_properties = {
        'args': ('--approve',),
        'kwargs': dict(help=description, nargs='+')
    }


class BrokerIdParameter(AbstractCommaSeparatedParameter):
    """brokerid=[id1,id2...]"""
    name = 'brokerid'
    description = 'Comma-separated and/or space-separated list of broker IDs'
    argparse_properties = {
        'args': ('brokers',),
        'kwargs': dict(help=description,
                       nargs='+')
    }


class ClientIdsParameter(AbstractCommaSeparatedParameter):
    """client_ids=[Set-of-ClientIdentity]"""
    name = 'client_ids'
    description = "The set of Client IDs by which to filter the user_tasks response"
    argparse_properties = {
        'args': ('--client-id', '--client-ids'),
        'kwargs': dict(help=description, nargs='+')
    }


class DestinationBrokerIdsParameter(AbstractCommaSeparatedParameter):
    """destination_broker_ids=[id1,id2...]"""
    name = 'destination_broker_ids'
    description = 'Comma-separated and/or space-separated list of broker IDs'
    argparse_properties = {
        'args': ('--destination-broker',
                 '--destination-brokers',
                 '--destination-broker-id',
                 '--destination-broker-ids'),
        'kwargs': dict(help=description,
                       nargs='+')
    }


class DiscardParameter(AbstractCommaSeparatedParameter):
    """discard=[id1,id2,...]"""
    name = 'discard'
    description = "The review IDs to discard"
    argparse_properties = {
        'args': ('--discard',),
        'kwargs': dict(help=description, nargs='+')
    }


class EndpointsParameter(AbstractCommaSeparatedParameter):
    """endpoints=[Set-of-{@link EndPoint}]"""
    name = 'endpoints'
    description = "The set of Endpoint by which to filter the user_tasks response"
    argparse_properties = {
        'args': ('--endpoint', '--endpoints'),
        'kwargs': dict(help=description, nargs='+')
    }


class GoalsParameter(AbstractCommaSeparatedParameter):
    """goals=[goal1,goal2...]"""
    name = 'goals'
    description = 'Comma-separated and/or space-separated ordered list of goals'
    argparse_properties = {
        'args': ('--goals',),
        'kwargs': dict(help=description, nargs='+')
    }


class ReplicaMovementStrategiesParameter(AbstractCommaSeparatedParameter):
    """replica_movement_strategies=[strategy1,strategy2...]"""
    name = 'replica_movement_strategies'
    description = 'Comma-separated and/or space-separated list of replica movement strategies'
    argparse_properties = {
        'args': ('--strategies',),
        'kwargs': dict(help=description, nargs='+')
    }


class ReviewIDsParameter(AbstractCommaSeparatedParameter):
    """review_ids=[id1,id2,...]"""
    name = 'review_ids'
    description = "The review IDs by which to filter the review_board response"
    argparse_properties = {
        'args': ('--review-ids', '--review-id', '--reviews', '--review'),
        'kwargs': dict(help=description, nargs='+')
    }


class TypesParameter(AbstractCommaSeparatedParameter):
    """types=[Set-of-{@link UserTaskManager.TaskState}]"""
    name = 'types'
    description = "The set of TaskStates by which to filter the user_tasks response"
    argparse_properties = {
        'args': ('--types', '--type', '--task-states', '--task-state'),
        'kwargs': dict(help=description, nargs='+')
    }


class UserTaskIdsParameter(AbstractCommaSeparatedParameter):
    """user_task_ids=[Set-of-USER-TASK-IDS]"""
    name = 'user_task_ids'
    description = "The set of UserTaskIDs by which to filter the user_tasks response"
    argparse_properties = {
        'args': ('--user-task-ids', '--user-task-id'),
        'kwargs': dict(help=description, nargs='+')
    }



# Copyright 2019 LinkedIn Corp. Licensed under the BSD 2-Clause License (the "License").
# See License in the project root for license information.

# Use the common Display library for showing non-final responses
from cruisecontrolclient.client.Display import display_response

# Use convenience function to redirect printing to stderr
from cruisecontrolclient.util.print import print_error

# To be able to make HTTP calls
import requests

# To allow us to make more-precise type hints
from typing import Callable, Dict  # noqa

# To be able to define a multithreaded way of interacting with cruise-control
from threading import Thread


class AbstractResponder(object):
    """
    This abstract class provides the skeleton for returning a final requests.Response
    object from cruise-control.

    This class should not be used directly, since the HTTP method is not defined
    in this class.

    See the children of AbstractJSONResponder and of AbstractTextResponder
    for the concrete classes that use GET and POST to retrieve a response
    from cruise-control.
    """

    def __init__(self, url: str, headers: Dict[str, str] = None):
        self.url: str = url
        self.headers: Dict[str, str] = headers

        # Open a session in order to handle cruise-control cookies
        self.session: requests.Session = requests.Session()

        # This abstract class does not define which of the session's
        # HTTP methods should be used
        self.session_http_method: Callable[[str], requests.Response] = None

    def retrieve_response(self) -> requests.Response:
        raise NotImplementedError


class AbstractTextResponder(AbstractResponder):
    """
    This abstract class provides the skeleton for returning a final requests.Response
    object from cruise-control where request.text is text-formatted.

    This class should not be used directly, since the HTTP method is not defined
    in this class.

    TextResponderGet and TextResponderPost are the concrete classes for
    using GET and POST to retrieve a response from cruise-control.
    """

    def __init__(self, url: str, headers: Dict[str, str] = None):
        if 'json=true' in url:
            raise ValueError("url must not contain the \"json=true\" parameter")
        AbstractResponder.__init__(self, url, headers)

    def retrieve_response(self) -> requests.Response:
        """
        Returns a final requests.Response object from cruise-control
        where Response.text is text-formatted.

        :return: requests.Response
        """
        # There's actually not a good way at present to do this with
        # a text response from cruise-control.
        #
        # It is much easier to determine whether every JSON response is "final".
        raise NotImplementedError


class AbstractJSONResponder(AbstractResponder):
    print_to_stdout_enabled = False
    """
    This abstract class provides the skeleton for returning a final requests.Response
    object from cruise-control where request.text is JSON-formatted.

    This class should not be used directly, since the HTTP method is not defined
    in this class.

    JSONResponderGet and JSONResponderPost are the concrete classes for
    using GET and POST to retrieve a response from cruise-control.

    This class does NOT display intermediate responses to humans.
    """

    def __init__(self, url: str, headers: Dict[str, str] = None):
        if 'json=true' not in url:
            raise ValueError("url must contain the \"json=true\" parameter")
        AbstractResponder.__init__(self, url, headers)
        # This abstract class _still_ does not define which HTTP method should be used

    def retrieve_response(self) -> requests.Response:
        """
        Returns a final requests.Response object from cruise-control
        where Response.text is JSON-formatted.

        :return: requests.Response
        """
        # cruise-control's JSON response has a 'progress' key in it so long
        # as the response is not final.
        #
        # Once the response is final, it does not contain the 'progress' key.
        #
        # Accordingly, keep getting the response from this session and checking
        # it for the 'progress' key.
        #
        # Return the response as JSON once we get a valid JSON response that we
        # think is 'final'.

        # Alert the humans to long-running poll
        if self.print_to_stdout_enabled:
            print_error("Starting long-running poll of {}".format(self.url))

        # TODO: Session.get and Session.post can return a plethora of exceptions;
        # they should be handled here.
        response = self.session_http_method(self.url, headers=self.headers)
        while 'progress' in response.json().keys():
            if self.print_to_stdout_enabled:
                display_response(response)
            response = self.session_http_method(self.url, headers=self.headers)

        # return the requests.response object
        return response


class AbstractJSONDisplayingResponder(AbstractJSONResponder):
    print_to_stdout_enabled = True
    """
    This class displays intermediate responses to humans via stdout.
    """


class JSONDisplayingResponderGet(AbstractJSONDisplayingResponder):
    """
    This class returns a final requests.Response object from cruise-control
    where Response.text is JSON-formatted, and where the HTTP method is GET.
    """

    def __init__(self, url: str, headers: Dict[str, str] = None):
        AbstractJSONDisplayingResponder.__init__(self, url, headers)
        self.session_http_method = self.session.get


class JSONDisplayingResponderPost(AbstractJSONDisplayingResponder):
    """
    This class returns a final requests.Response object from cruise-control
    where Response.text is JSON-formatted, and where the HTTP method is POST.
    """

    def __init__(self, url: str, headers: Dict[str, str] = None):
        AbstractJSONDisplayingResponder.__init__(self, url, headers)
        self.session_http_method = self.session.post


class AbstractJSONResponderThread(Thread):
    """
    This abstract class defines a Thread whose purpose is to communicate with
    cruise-control and return a requests.Response object.

    This class should not be used directly, since the HTTP method is not defined
    in this class.

    JSONResponderGetThread and JSONResponderPostThread are the concrete classes
    for using GET and POST to retrieve a response from cruise-control.
    """

    def __init__(self):
        Thread.__init__(self)
        # This abstract class does not define which concrete JSONResponder class to use
        self.json_responder = None
        self.response = None

    def run(self):
        self.response = self.json_responder.retrieve_response()

    def get_response(self):
        return self.response

    def get_json_response(self):
        return self.response.json()


class JSONResponderGetThread(AbstractJSONResponderThread):
    """
    This class defines a Thread whose purpose is to communicate with cruise-control
    via HTTP GET and return a requests.Response object.
    """

    def __init__(self, url: str, headers: Dict[str, str] = None):
        AbstractJSONResponderThread.__init__(self)
        self.json_responder = JSONDisplayingResponderGet(url, headers)


class JSONResponderPostThread(AbstractJSONResponderThread):
    """
    This class defines a Thread whose purpose is to communicate with cruise-control
    via HTTP POST and return a requests.Response object.
    """

    def __init__(self, url: str, headers: Dict[str, str] = None):
        AbstractJSONResponderThread.__init__(self)
        self.json_responder = JSONDisplayingResponderPost(url, headers)



# Copyright 2019 LinkedIn Corp. Licensed under the BSD 2-Clause License (the "License").
# See License in the project root for license information.

# This is a package.
#
# Please do not add the code to this file unless absolutely
# necessary because __init__.py files are not needed with Python 3
